/* -*- tab-width:8; coding:utf-8 -*-
 *
 * Set / inspect region settings on a DVD drive. This is most interesting
 * on a RPC Phase-2 drive, of course.
 *
 * Usage: dvd_region [ -d device ] [ -s -r region ]
 *
 * Jens Axboe <axboe@suse.de>
 *
 */
/* 
 * Some slight adaptations to make the program work under Solaris 10.
 * Kudos to the libdvdcss people.
 *
 * Warning: I haven't tested this a lot (just don't have enough
 * DVD drives...), so it might still contain some serious bugs. Use at
 * your own risk.
 * 
 * Raimund Steger <rs@mytum.de>, 2010-11-15
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#if defined(__linux__)

#include <linux/cdrom.h>
#define DEFAULT_DEVICE "/dev/dvd"

#elif defined (__SVR4) && defined (__sun)

/* Includes code from libdvdcss: -- rs */
#include "solaris_uscsi.h"

/* That's the vold path. Should be fine at least 
 * with Solaris 10. -- rs */
#define DEFAULT_DEVICE "/vol/dev/aliases/cdrom0"

#else
#error No supported operating system found
#endif

int do_region(int fd,int *p_type,int *p_mask,int *p_scheme,
              int *p_vra,int *p_ucca,int set,int region)
{
    int ret;

#if defined(__linux__)
    dvd_authinfo auth_info;
    memset(&auth_info,0,sizeof(auth_info));
    auth_info.type = DVD_LU_SEND_RPC_STATE;

    if((ret = ioctl(fd,DVD_AUTH,&auth_info)) < 0)
    {
        perror("do_region get");
        return ret;
    }

    *p_type = auth_info.lrpcs.type;
    *p_vra = auth_info.lrpcs.vra;
    *p_ucca = auth_info.lrpcs.ucca;
    *p_mask = auth_info.lrpcs.region_mask;
    *p_scheme = auth_info.lrpcs.rpc_scheme;
#else
    INIT_USCSI(GPCMD_REPORT_KEY,8);

    rs_cdb.cdb_opaque[10] = DVD_REPORT_RPC;

    ret = SolarisSendUSCSI(fd,&sc);

    if(ret < 0 || sc.uscsi_status )
    {
        perror("do_region get");
        ret = -1;
        return ret;    
    }

    *p_type = p_buffer[4] >> 6;
    *p_vra = (p_buffer[4] >> 3) & 0x7;
    *p_ucca = p_buffer[4] & 0x7;
    *p_mask = p_buffer[5];
    *p_scheme = p_buffer[6];
#endif

    if(set) 
    {
        if(region > 8 || region <= 0) 
        {
            printf("Invalid region (%d)\n", region);
            return 1;
        }
        printf("Setting drive region can only be done a finite " \
               "number of times, press CTRL-C now to cancel!\n");
        getchar();

#if defined(__linux__)
        memset(&auth_info,0,sizeof(auth_info));
        auth_info.type = DVD_HOST_SEND_RPC_STATE;
        auth_info.hrpcs.pdrc = ~(1 << (region - 1));
        printf("Setting region to %d (0x%x)\n",region,auth_info.hrpcs.pdrc);
        if((ret = ioctl(fd,DVD_AUTH,&auth_info)) < 0) 
        {
            perror("do_region set");
            return ret;
        }
#else
        INIT_USCSI(GPCMD_SEND_KEY,8);

        rs_cdb.cdb_opaque[10] = DVD_SEND_RPC;

        p_buffer[1] = 6;
        p_buffer[4] = ~(1 << (region - 1));
        printf("Setting region to %d (%x)\n",region,p_buffer[4]);
        ret = SolarisSendUSCSI(fd,&sc);
        if(ret < 0 || sc.uscsi_status)
        {
            perror("do_region set");
            ret = -1;
            return ret;        
        }

#endif
    }
    return 0;
}

void print_region(int p_type,int p_mask,int p_scheme,int p_vra,int p_ucca)
{
    printf("Drive region info:\n");

    printf("Type: ");
    switch(p_type) 
    {
    case 0:
	printf("No drive region setting\n");
	break;
    case 1:
	printf("Drive region is set\n");
	break;
    case 2:
	printf("Drive region is set, with additional "		\
	       "restrictions required to make a change\n");
	break;
    case 3:
	printf("Drive region has been set permanently, but "		\
	       "may be reset by the vendor if necessary\n");
	break;
    default:
	printf("Invalid (%x)\n",p_type);
	break;
    }

    printf("%d vendor resets available\n",p_vra);
    printf("%d user controlled changes available\n",p_ucca);
    printf("Region: ");
    if(p_mask)
    {
        printf("%d playable\n", ffs(~p_mask));
    }
    else
    {
        printf("non-rpc (all)\n");
    }

    printf("RPC Scheme: ");
    switch(p_scheme) 
    {
    case 0:
	printf("The Logical Unit does not enforce Region "	\
	       "Playback Controls (RPC)\n");
	break;
    case 1:
	printf("The Logical Unit _shall_ adhere to the "
	       "specification and all requirements of the "	\
	       "CSS license agreement concerning RPC\n");
	break;
    default:
	printf("Reserved (%x)\n",p_scheme);
    }
}

void usage()
{
    printf("Set / inspect region settings on a DVD drive.\n");
    printf("Usage: dvd_region [ -d device ] [ -s -r region ]\n");
}

int main(int argc,char *argv[])
{
    char device_name[FILENAME_MAX],c,set,get,region;

    int fd,ret;
    int p_type,p_mask,p_scheme,p_vra,p_ucca;

    strcpy(device_name,DEFAULT_DEVICE);
    set = 0;
    get = 1;
    region = -1;
    while((c = getopt(argc, argv, "d:shr:")) != EOF) 
    {
        switch(c) 
        {
	case 'd':
	    strncpy(device_name,optarg,FILENAME_MAX - 1);
	    break;
	case 's':
	    set = 1;
	    break;
	case 'r':
	    region = strtoul(optarg,NULL,10);
	    printf("Region to set: %d\n",region);
	    break;
	case 'h':
	    usage();
	    return 0;
	}
    }

    if(set&&region<0)
    {
	fprintf(stderr,"No region or invalid region specified.\n");
	return 1;
    }

    printf("Using device: %s\n",device_name);
    
    if((fd = open(device_name, O_RDONLY | O_NONBLOCK)) < 0) 
    {
        perror("open");
        return 1;
    }

    if((ret = do_region(fd,&p_type,&p_mask,&p_scheme,&p_vra,&p_ucca,0,0)))
    {
        return ret;
    }

    print_region(p_type,p_mask,p_scheme,p_vra,p_ucca);

    if(set)
    {
        if((ret = do_region(fd,&p_type,&p_mask,&p_scheme,
                               &p_vra,&p_ucca,1,region)) < 0)
        {
            return ret;
        }
    }

    return 0;
}
